#pragma once

#include "GameObject.h"


class Player : public GameObject
{
  public:

    float         AnimDelay;
    float         SizeFactor;
    bool          Moving;
    int           LastDir;



    Player();


    virtual void Display();
    virtual void Update( Level& Level, float ElapsedTime );
    virtual RECT Bounds();

    void Grow( Level& Level, float GrowFactor );
    void Shrink( Level& Level, float GrowFactor );

    virtual void OnObjectRemoved( Level& Level, GameObject* pKilledObject );

};

